# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import json
import logging
from collections import OrderedDict

from liota.dccs.dcc import DataCenterComponent
from liota.entities.registered_entity import RegisteredEntity
from liota.entities.edge_systems.edge_system import EdgeSystem
from liota.entities.devices.device import Device
from liota.entities.metrics.metric import Metric
from liota.entities.metrics.registered_metric import RegisteredMetric
from liota.lib.utilities.si_unit import parse_unit, UnsupportedUnitError

log = logging.getLogger(__name__)


class AWSIoT(DataCenterComponent):
    """
    DCC for AWSIoT Platform.
    """
    def __init__(self, con, enclose_metadata=False):
        """
        :param con: DccComms Object
        :param enclose_metadata: Include Gateway, Device and Metric names as part of payload or not
        """
        super(AWSIoT, self).__init__(
            comms=con
        )
        self.enclose_metadata = enclose_metadata

    def register(self, entity_obj):
        """
        :param entity_obj: Entity Object
        :return: RegisteredEntity Object
        """
        log.info("Registering resource with AWSIoT DCC {0}".format(entity_obj.name))
        super(AWSIoT, self).register(entity_obj)
        if isinstance(entity_obj, Metric):
            return RegisteredMetric(entity_obj, self, None)
        else:
            return RegisteredEntity(entity_obj, self, None)

    def create_relationship(self, reg_entity_parent, reg_entity_child):
        """
        This method creates Parent-Child relationship.  Supported relationships are:

               EdgeSystem
                   |                                      EdgeSystem
                Device                   (or)                |
                   |                                    RegisteredMetric
             RegisteredMetric

        However, A single EdgeSystem can have multiple child Devices and a each Device can have
        multiple child Metrics.

        :param reg_entity_parent: Registered EdgeSystem or Registered Device Object
        :param reg_entity_child:  Registered Device or Registered Metric Object
        :return: None
        """
        # Validating here helps in enclose_metadata
        if not isinstance(reg_entity_parent.ref_entity, EdgeSystem) and \
                not isinstance(reg_entity_parent.ref_entity, Device):
            log.error("reg_entity_parent should either be a Registered EdgeSystem or a Device")
            raise TypeError("reg_entity_parent should either be a Registered EdgeSystem or a Device")

        if not isinstance(reg_entity_child.ref_entity, Device) and \
                not isinstance(reg_entity_child, RegisteredMetric):
            log.error("reg_entity_child should either be a Registered Device or Metric")
            raise TypeError("reg_entity_child should either be a Registered Device or Metric")

        reg_entity_child.parent = reg_entity_parent

    def _get_entity_hierarchy(self, reg_entity):
        """
        NOTE: This method should be moved to utility if other DCCs require this functionality.
        :param reg_entity: RegisteredMetric Object
        :return: A list with entity names
                 - [edge_system_name, device_name, metric_name] (or)
                 - [edge_system_name, metric_name]
        """
        if not isinstance(reg_entity, RegisteredEntity):
            log.error("RegisteredEntity is expected")
            raise TypeError("RegisteredEntity is expected")

        def extract_hierarchy(reg_entity):
            """
            Recursive function to get entity names
            :param reg_entity: RegisteredEntity Object
            :return:
            """
            if reg_entity is None:
                return []
            return extract_hierarchy(reg_entity.parent) + [reg_entity.ref_entity.name]

        return extract_hierarchy(reg_entity)

    def _format_data(self, reg_metric):
        """
        :param reg_metric: Registered Metric Object
        :return: Payload in JSON format
        """
        met_cnt = reg_metric.values.qsize()
        if 0 == met_cnt:
            return

        _list = []
        for _ in range(met_cnt):
            m = reg_metric.values.get(block=True)
            if m is not None:
                _list.append(OrderedDict([('value', m[1]), ('timestamp', m[0])]))

        payload = OrderedDict()
        if self.enclose_metadata:
            _entity_hierarchy = self._get_entity_hierarchy(reg_metric)
            #  EdgeSystem and Device's name will be added with payload
            if len(_entity_hierarchy) == 3:
                payload['edge_system_name'] = _entity_hierarchy[0]
                payload['device_name'] = _entity_hierarchy[1]
            # EdgeSystem's name will be added with payload
            elif len(_entity_hierarchy) == 2:
                payload['edge_system_name'] = _entity_hierarchy[0]
            else:
                # Not raising error.
                # Metrics can be published even if error occurred while
                # constructing payload for enclose_metadata
                log.error("Error occurred while constructing payload")
        payload['metric_name'] = reg_metric.ref_entity.name
        payload['metric_data'] = [_ for _ in _list]
        # TODO: Make this as part of si_unit.py
        # Handling Base, Derived and Prefixed Units
        if reg_metric.ref_entity.unit is None:
            payload['unit'] = 'null'
        else:
            try:
                unit_tuple = parse_unit(reg_metric.ref_entity.unit)
                if unit_tuple[0] is None:
                    # Base and Derived Units
                    payload['unit'] = unit_tuple[1]
                else:
                    # Prefixed or non-SI Units
                    payload['unit'] = unit_tuple[0] + unit_tuple[1]
            except UnsupportedUnitError as err:
                # Not raising error.
                # Metrics can be published even if unit is unsupported
                payload['unit'] = 'null'
                log.error(str(err))
        return json.dumps(payload)

    def set_properties(self, reg_entity, properties):
        raise NotImplementedError

    def unregister(self, entity_obj):
        raise NotImplementedError



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import logging
import os
import ssl
import sys
import time

import paho.mqtt.client as paho

from liota.lib.utilities.utility import systemUUID


log = logging.getLogger(__name__)


class Mqtt():
    """
    MQTT Transport implementation for LIOTA. It internally uses Python Paho library.
    """

    def on_connect(self, client, userdata, flags, rc):
        """
        Invoked on successful connection to a broker after connection request

        :param client: The client instance for this callback
        :param userdata: The private user data as set in Client() or userdata_set()
        :param flags: Response flags sent by the broker
        :param rc: The connection result
        :return:
        """
        self._connect_result_code = rc
        self._disconnect_result_code = sys.maxsize
        log.info("Connected with result code : {0} : {1} ".format(str(rc), paho.connack_string(rc)))

    def on_disconnect(self, client, userdata, rc):
        """
        Invoked when disconnected from broker.  Two scenarios are possible:

        1) Broker rejects a connection request
        2) Client initiated disconnect

        :param client: The client instance for this callback
        :param userdata: The private user data as set in Client() or userdata_set()
        :param rc: The connection result
        :return:
        """
        self._connect_result_code = sys.maxsize
        self._disconnect_result_code = rc
        log.info("Disconnected with result code : {0} : {1} ".format(str(rc), paho.connack_string(rc)))

    def on_message(self, client, userdata, msg):
        """
        Invoked when message received on a subscribed topic.

        :param client: The client instance for this callback
        :param userdata: The private user data as set in Client() or userdata_set()
        :param msg: An instance of MQTTMessage. This is a class with members topic, payload, qos, retain.
        :return:
        """
        log.debug("On Message {0} {1} {2}".format(msg.topic, str(msg.qos), str(msg.payload)))

    def on_publish(self, client, userdata, mid):
        """
        Called when publish transmission is completed.

        :param client: The client instance for this callback
        :param userdata: The private user data as set in Client() or userdata_set()
        :param mid: Message ID
        :return:
        """
        log.debug("mid: {0}".format(str(mid)))

    def on_subscribe(self, client, userdata, mid, granted_qos):
        """
        Invoked when the broker responds to subscribe request.

        :param client: The client instance for this callback
        :param userdata: The private user data as set in Client() or userdata_set()
        :param mid: Message ID
        :param granted_qos: Greanted QoS by the broker
        :return:
        """
        log.debug("Subscribed: {0} {1}".format(str(mid), str(granted_qos)))

    def on_unsubscribe(self, client, userdata, mid):
        """
        Invoked when the broker responds to an unsubscribe request.

        :param client: The client instance for this callback
        :param userdata: The private user data as set in Client() or userdata_set()
        :param mid: Message ID
        :return:
        """
        log.debug("Unsubscribed: {0}".format(str(mid)))

    def __init__(self, url, port, identity=None, tls_conf=None, qos_details=None, client_id="",
                 clean_session=False, userdata=None, protocol="MQTTv311", transport="tcp", keep_alive=60,
                 enable_authentication=False, conn_disconn_timeout=10):

        """
        :param url: MQTT Broker URL or IP
        :param port: MQTT Broker Port
        :param identity: Identity Object
        :param tls_conf: TLSConf object
        :param qos_details: QoSDetails object
        :param client_id: Client ID
        :param clean_session: Connect with Clean session or not
        :param userdata: userdata is user defined data of any type that is passed as the "userdata"
                         parameter to callbacks.

        :param protocol: allows explicit setting of the MQTT version to use for this client
        :param transport: Set transport to "websockets" to use WebSockets as the transport
                          mechanism. Set to "tcp" to use raw TCP, which is the default.

        :param keep_alive: KeepAliveInterval
        :param enable_authentication: Enable user-name password authentication or not
        :param username: Username for authentication
        :param password: Password for authentication
        :param conn_disconn_timeout: Connect-Disconnect-Timeout
        """
        self.url = url
        self.port = port
        self.identity = identity
        self.tls_conf = tls_conf
        self.qos_details = qos_details
        self.client_id = client_id
        self.clean_session = clean_session
        self.userdata = userdata
        self.protocol = protocol
        self.transport = transport
        self.keep_alive = keep_alive
        self.enable_authentication = enable_authentication
        self._conn_disconn_timeout = conn_disconn_timeout
        if self.clean_session:
            # If user passes client_id, it'll be used.  Otherwise, it is left to the underlying paho
            # to generate random client_id
            self._paho_client = paho.Client(self.client_id, self.clean_session, self.userdata,
                                            protocol=getattr(paho, self.protocol), transport=self.transport)
            log.info("clean_session is set to True")
        else:
            #  client_id is either auto-generated or provided by user
            self._paho_client = paho.Client(self.client_id, self.clean_session, self.userdata,
                                            protocol=getattr(paho, self.protocol), transport=self.transport)
            log.info("clean_session is set to False")

        self._connect_result_code = sys.maxsize
        self._disconnect_result_code = sys.maxsize
        self._paho_client.on_message = self.on_message
        self._paho_client.on_publish = self.on_publish
        self._paho_client.on_subscribe = self.on_subscribe
        self._paho_client.on_connect = self.on_connect
        self._paho_client.on_disconnect = self.on_disconnect
        self.connect_soc()

    def connect_soc(self):
        """
        Establishes connection with MQTT Broker
        :return:
        """
        # Set up TLS support
        if self.tls_conf:

            # Validate CA certificate path
            if self.identity.root_ca_cert:
                if not(os.path.exists(self.identity.root_ca_cert)):
                    log.error("Error : Wrong CA certificate path.")
                    raise ValueError("Error : Wrong CA certificate path.")
            else:
                log.error("Error : CA certificate path is missing")
                raise ValueError("Error : CA certificate path is missing")

            # Validate client certificate path
            if self.identity.cert_file:
                if os.path.exists(self.identity.cert_file):
                    client_cert_available = True
                else:
                    log.error("Error : Wrong client certificate path.")
                    raise ValueError("Error : Wrong client certificate path.")
            else:
                client_cert_available = False

            # Validate client key file path
            if self.identity.key_file:
                if os.path.exists(self.identity.key_file):
                    client_key_available = True
                else:
                    log.error("Error : Wrong client key path.")
                    raise ValueError("Error : Wrong client key path.")
            else:
                client_key_available = False

            '''
                Multiple conditions for certificate validations
                # 1. Both Client certificate and key file should be present
                # 2. If both are not there proceed without client certificate and key
                # 3. If client certificate is not there throw an error
                # 4. If client key is not there throw an error
            '''

            if client_cert_available and client_key_available:
                log.debug("Certificates : ", self.identity.root_ca_cert, self.identity.cert_file,
                          self.identity.key_file)

                self._paho_client.tls_set(self.identity.root_ca_cert, self.identity.cert_file,
                                          self.identity.key_file,
                                          cert_reqs=getattr(ssl, self.tls_conf.cert_required),
                                          tls_version=getattr(ssl, self.tls_conf.tls_version),
                                          ciphers=self.tls_conf.cipher)
            elif not client_cert_available and not client_key_available:
                self._paho_client.tls_set(self.identity.root_ca_cert,
                                          cert_reqs=getattr(ssl, self.tls_conf.cert_required),
                                          tls_version=getattr(ssl, self.tls_conf.tls_version),
                                          ciphers=self.tls_conf.cipher)
            elif not client_cert_available and client_key_available:
                log.error("Error : Client key found, but client certificate not found")
                raise ValueError("Error : Client key found, but client certificate not found")
            else:
                log.error("Error : Client key found, but client certificate not found")
                raise ValueError("Error : Client certificate found, but client key not found")
            log.info("TLS support is set up.")

        # Set up username-password
        if self.enable_authentication:
            if not self.identity.username:
                log.error("Username not found")
                raise ValueError("Username not found")
            elif not self.identity.password:
                log.error("Password not found")
                raise ValueError("Password not found")
            else:
                self._paho_client.username_pw_set(self.identity.username, self.identity.password)

        if self.qos_details:
            # Set QoS parameters
            self._paho_client.max_inflight_messages_set(self.qos_details.in_flight)
            self._paho_client.max_queued_messages_set(self.qos_details.queue_size)
            self._paho_client.message_retry_set(self.qos_details.retry)

        # Connect with MQTT Broker
        self._paho_client.connect(host=self.url, port=self.port, keepalive=self.keep_alive)

        # Start network loop to handle auto-reconnect
        self._paho_client.loop_start()
        ten_ms_count = 0
        while (ten_ms_count != self._conn_disconn_timeout * 100) and (self._connect_result_code == sys.maxsize):
            ten_ms_count += 1
            time.sleep(0.01)
        if self._connect_result_code == sys.maxsize:
            log.error("Connection timeout.")
            #  Stopping background network loop as connection establishment failed.
            self._paho_client.loop_stop()
            raise Exception("Connection Timeout")
        elif self._connect_result_code == 0:
            log.info("Connected to MQTT Broker.")
            log.info("Connect time consumption: " + str(float(ten_ms_count) * 10) + "ms.")
        else:
            log.error("Connection error with result code : {0} : {1} ".
                      format(str(self._connect_result_code), paho.connack_string(self._connect_result_code)))
            #  Stopping background network loop as connection establishment failed.
            self._paho_client.loop_stop()
            raise Exception("Connection error with result code : {0} : {1} ".
                      format(str(self._connect_result_code), paho.connack_string(self._connect_result_code)))

    def publish(self, topic, message, qos, retain=False):
        """
        Publishes message to the MQTT Broker

        :param topic: Publish topic
        :param message: Message to be published
        :param qos: Publish QoS
        :param retain: Message to be retained or not
        :return:
        """
        # TODO: Retry logic to be designed
        try:
            mess_info = self._paho_client.publish(topic, message, qos, retain)
            log.info("Publishing Message ID : {0} with result code : {1} ".format(mess_info.mid, mess_info.rc))
            log.debug("Published Topic:{0}, Payload:{1}, QoS:{2}".format(topic, message, qos))
        except Exception:
            log.exception("MQTT Publish exception traceback..")

    def subscribe(self, topic, qos, callback):
        """
        Subscribes to a topic with given callback

        :param topic: Subscribe topic
        :param qos: Subscribe QoS
        :param callback:  Callback for the topic
        :return:
        """
        try:
            subscribe_response = self._paho_client.subscribe(topic, qos)
            self._paho_client.message_callback_add(topic, callback)
            log.info("Topic subscribed with information: " + str(subscribe_response))
        except Exception:
            log.exception("MQTT subscribe exception traceback..")

    def disconnect(self):
        """
        Disconnects from MQTT Broker
        :return:
        """
        self._paho_client.disconnect()
        ten_ms_count = 0
        while (ten_ms_count != self._conn_disconn_timeout * 100) and (self._disconnect_result_code == sys.maxsize):
            ten_ms_count += 1
            time.sleep(0.01)
        if self._disconnect_result_code == sys.maxsize:
            log.error("Disconnect timeout.")
            raise Exception("Disconnection Timeout")
        elif self._disconnect_result_code == 0:
            log.info("Disconnected from MQTT Broker.")
            log.info("Disconnect time consumption: " + str(float(ten_ms_count) * 10) + "ms.")
            #  Disconnect is successful.  Stopping background network loop.
            self._paho_client.loop_stop()
        else:
            log.error("Disconnect error with result code : {0} : {1} ".
                      format(str(self._disconnect_result_code), paho.connack_string(self._disconnect_result_code)))
            raise Exception("Disconnect error with result code : {0} : {1} ".
                      format(str(self._disconnect_result_code), paho.connack_string(self._disconnect_result_code)))

    def get_client_id(self):
        """
        Returns client-id
        :return:
        """
        return self._paho_client._client_id


class QoSDetails:
    """
    Encapsulates config parameters related to Quality of Service
    """
    def __init__(self, in_flight, queue_size, retry):
        """
        :param in_flight: Set maximum no. of messages with QoS>0 that can be part way
                          through their network flow at once. Default is 20.
        :param queue_size: Set the maximum number of messages in the outgoing message queue. 0 means unlimited.
        :param retry: Set the timeout in seconds before a message with QoS>0 is retried. 20 seconds by default.
        """
        self.in_flight = in_flight
        self.queue_size = queue_size
        self.retry = retry


class MqttMessagingAttributes:
    """
    Encapsulates MessagingAttributes related to MQTT.

     This class enables the following options for developers in LIOTA
     ----------------------------------------------------------------

     a) Use single publish and subscribe topic generated by LIOTA for an EdgeSystem, its Devices and its Metrics.
        - Publish topic for all Metrics will be 'liota/generated_local_uuid_of_edge_system/request'
        - Subscribe topic will be 'liota/generated_local_uuid_of_edge_system/response'
     b) Use custom single publish and subscribe topic for an EdgeSystem, its Devices and Metrics.

     - In the above two cases, MQTT message's payload MUST be self-descriptive so that subscriber can subscribe
      process accordingly to a single topic by parsing payload.

     c) Use custom publish and subscribe topics for Metrics.
     - In this case, MQTT message's payload need not be self-descriptive.  Subscribers can subscribe to
     appropriate topics accordingly.

     d) Use combination of (a) and (c) or (b) and (c).

    """
    def __init__(self, edge_system_name=None, pub_topic=None, sub_topic=None, pub_qos=1, sub_qos=1, pub_retain=False,
                 sub_callback=None):
        """
        :param edge_system_name: Name of the EdgeSystem
        :param pub_topic: Publish topic of EdgeSystem or RegisteredMetric
        :param sub_topic: Subscribe topic of EdgeSystem or RegisteredMetric
        :param pub_qos: Publish QoS
        :param sub_qos: Subscribe QoS
        :param pub_retain: Publish Retain Flag
        :param sub_callback: Subscribe Callback
        """
        if edge_system_name:
            #  For Project ICE and Non-Project ICE, topics will be auto-generated if edge_system_name is not None
            self.pub_topic = 'liota/' + systemUUID().get_uuid(edge_system_name) + '/request'
            self.sub_topic = 'liota/' + systemUUID().get_uuid(edge_system_name) + '/response'
        else:
            #  When edge_system_name is None, pub_topic or sub_topic must be provided
            self.pub_topic = pub_topic
            self.sub_topic = sub_topic

        #  General validation
        if pub_qos not in range(0, 3) or sub_qos not in range(0, 3):
            log.error("QoS should either be 0 or 1 or 2")
            raise ValueError("QoS should either be 0 or 1 or 2")
        if not isinstance(pub_retain, bool):
            log.error("pub_retain must be a boolean")
            raise ValueError("pub_retain must be a boolean")
        if sub_callback is not None:
            if not callable(sub_callback):
                log.error("sub_callback should either be None or callable")
                raise ValueError("sub_callback should either be None or callable")

        log.info("Pub Topic is:{0}".format(self.pub_topic))
        log.info("Sub Topic is:{0}".format(self.sub_topic))
        self.pub_qos = pub_qos
        self.sub_qos = sub_qos
        self.pub_retain = pub_retain
        self.sub_callback = sub_callback



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import json
import logging
import time
import threading
import ConfigParser
import os
import Queue
from time import gmtime, strftime
from threading import Lock
import xml.etree.cElementTree as ET
from xml.dom import minidom

from liota.dccs.dcc import DataCenterComponent, RegistrationFailure
from liota.lib.protocols.helix_protocol import HelixProtocol
from liota.entities.metrics.metric import Metric
from liota.lib.utilities.utility import LiotaConfigPath, getUTCmillis, mkdir, read_liota_config, store_edge_system_uuid
from liota.lib.utilities.si_unit import parse_unit
from liota.entities.metrics.registered_metric import RegisteredMetric
from liota.entities.registered_entity import RegisteredEntity

log = logging.getLogger(__name__)


class IotControlCenter(DataCenterComponent):
    """ The implementation of IoTCC cloud provider solution

    """

    def __init__(self, username, password, con):
        log.info("Logging into DCC")
        self.comms = con
        self.username = username
        self.password = password
        thread = threading.Thread(target=self.comms.receive)
        thread.daemon = True
        # This thread will continuously run in background to receive response or actions from DCC
        thread.start()
        self.proto = HelixProtocol(self.comms, username, password)
        self._iotcc_json = self._create_iotcc_json()
        self.counter = 0
        self.recv_msg_queue = self.comms.userdata
        self.dev_file_path = self._get_file_storage_path("dev_file_path")
        # Liota internal entity file system path special for iotcc
        self.entity_file_path = self._get_file_storage_path("entity_file_path")
        self.file_ops_lock = Lock()

        def on_response(msg):
            try:
                log.debug("Received msg: {0}".format(msg))
                json_msg = json.loads(msg)
                self.proto.on_receive(json_msg)
                if json_msg["type"] == "connection_response" and json_msg["body"]["result"] == "succeeded":
                    log.info("Connection verified")
                    return True
                else:
                    log.debug("Processed msg: {0}".format(json_msg["type"]))
                    on_response(self.recv_msg_queue.get(True,300))
            except Exception as error:
                log.error("HelixProtocolException: " + repr(error))

        # Block on Queue for not more then 300 seconds else it will raise an exception
        on_response(self.recv_msg_queue.get(True,300))
        log.info("Logged in to DCC successfully")

    def register(self, entity_obj):
        """ Register the objects

        """
        if isinstance(entity_obj, Metric):
            # reg_entity_id should be parent's one: not known here yet
            # will add in creat_relationship(); publish_unit should be done inside
            return RegisteredMetric(entity_obj, self, None)
        else:
            # finally will create a RegisteredEntity
            log.info("Registering resource with IoTCC {0}".format(entity_obj.name))

            def on_response(msg):
                try:
                    log.debug("Received msg: {0}".format(msg))
                    json_msg = json.loads(msg)
                    log.debug("Processed msg: {0}".format(json_msg["type"]))
                    if json_msg["type"] == "create_or_find_resource_response" and json_msg["body"]["uuid"] != "null" and \
                                    json_msg["body"]["id"] == entity_obj.entity_id:
                        log.info("FOUND RESOURCE: {0}".format(json_msg["body"]["uuid"]))
                        self.reg_entity_id = json_msg["body"]["uuid"]
                    else:
                        log.info("Waiting for resource creation")
                        on_response(self.recv_msg_queue.get(True,300))
                except:
                    raise Exception("Exception while registering resource")

            if entity_obj.entity_type == "EdgeSystem":
                entity_obj.entity_type = "HelixGateway"
            self.comms.send(json.dumps(
                self._registration(self.next_id(), entity_obj.entity_id, entity_obj.name, entity_obj.entity_type)))
            on_response(self.recv_msg_queue.get(True,300))
            if not self.reg_entity_id:
                raise RegistrationFailure()
            log.info("Resource Registered {0}".format(entity_obj.name))
            if entity_obj.entity_type == "HelixGateway":
                self.store_reg_entity_details(entity_obj.entity_type, entity_obj.name, self.reg_entity_id)
                store_edge_system_uuid(entity_name=entity_obj.name, entity_id=entity_obj.entity_id,
                                       reg_entity_id=self.reg_entity_id)
                with self.file_ops_lock:
                    self.store_reg_entity_attributes("EdgeSystem", entity_obj.name,
                                                     self.reg_entity_id, None, None)
            else:
                self.store_reg_entity_details(entity_obj.entity_type, entity_obj.name, self.reg_entity_id)
                # get dev_type, and prop_dict if possible
                with self.file_ops_lock:
                    self.store_reg_entity_attributes("Devices", entity_obj.name, self.reg_entity_id,
                                                     entity_obj.entity_type, None)

            return RegisteredEntity(entity_obj, self, self.reg_entity_id)

    def unregister(self, entity_obj):
        """ Unregister the objects
        """
        log.info("Unregistering resource with IoTCC {0}".format(entity_obj.ref_entity.name))

        def on_response(msg):
            try:
                log.debug("Received msg: {0}".format(msg))
                json_msg = json.loads(msg)
                log.debug("Processed msg: {0}".format(json_msg["type"]))
                if json_msg["type"] == "remove_resource_response" and json_msg["body"]["result"] == "succeeded":
                        log.info("Unregistration of resource {0} with IoTCC succeeded".format(entity_obj.ref_entity.name))
                else:
                        log.info("Unregistration of resource {0} with IoTCC failed".format(entity_obj.ref_entity.name))
            except:
                raise Exception("Exception while unregistering resource")

        self.comms.send(json.dumps(self._unregistration(self.next_id(), entity_obj.reg_entity_id)))
        on_response(self.recv_msg_queue.get(True,20))
        self.remove_reg_entity_details(entity_obj.ref_entity.name, entity_obj.reg_entity_id)
        if entity_obj.ref_entity.entity_type != "HelixGateway":
            self.store_device_info(entity_obj.reg_entity_id, entity_obj.ref_entity.name, entity_obj.ref_entity.entity_type, None, True)
        log.info("Unregistration of resource {0} with IoTCC complete".format(entity_obj.ref_entity.name))

    def create_relationship(self, reg_entity_parent, reg_entity_child):
        """ This function initializes all relations between Registered Entities.

            Parameters:
            - obj: The object that has just obtained an UUID
        """
        # sanity check: must be RegisteredEntity or RegisteredMetricRegisteredMetric
        if (not isinstance(reg_entity_parent, RegisteredEntity) \
                    and not isinstance(reg_entity_parent, RegisteredMetric)) \
                or (not isinstance(reg_entity_child, RegisteredEntity) \
                            and not isinstance(reg_entity_child, RegisteredMetric)):
            raise TypeError()

        reg_entity_child.parent = reg_entity_parent
        if isinstance(reg_entity_child, RegisteredMetric):
            # should save parent's reg_entity_id
            reg_entity_child.reg_entity_id = reg_entity_parent.reg_entity_id
            entity_obj = reg_entity_child.ref_entity
            self.publish_unit(reg_entity_child, entity_obj.name, entity_obj.unit)
        else:
            self.comms.send(json.dumps(self._relationship(self.next_id(),
                                                          reg_entity_parent.reg_entity_id,
                                                          reg_entity_child.reg_entity_id)))

    def _registration(self, msg_id, res_id, res_name, res_kind):
        return {
            "transactionID": msg_id,
            "type": "create_or_find_resource_request",
            "body": {
                "kind": res_kind,
                "id": res_id,
                "name": res_name
            }
        }

    def _relationship(self, msg_id, parent_res_uuid, child_res_uuid):
        return {
            "transactionID": msg_id,
            "type": "create_relationship_request",
            "body": {
                "parent": parent_res_uuid,
                "child": child_res_uuid
            }
        }

    def _properties(self, msg_id, res_uuid, res_kind, timestamp, properties):
        msg = {
            "transationID": msg_id,
            "type": "add_properties",
            "uuid": res_uuid,
            "body": {
                "kind": res_kind,
                "timestamp": timestamp,
                "property_data": []
            }
        }
        for key, value in properties.items():
            msg["body"]["property_data"].append({"propertyKey": key, "propertyValue": value})
        return msg

    def _format_data(self, reg_metric):
        met_cnt = reg_metric.values.qsize()
        if met_cnt == 0:
            return
        _timestamps = []
        _values = []
        for _ in range(met_cnt):
            m = reg_metric.values.get(block=True)
            if m is not None:
                _timestamps.append(m[0])
                _values.append(m[1])
        if _timestamps == []:
            return
        return json.dumps({
            "type": "add_stats",
            "uuid": reg_metric.reg_entity_id,
            "metric_data": [{
                "statKey": reg_metric.ref_entity.name,
                "timestamps": _timestamps,
                "data": _values
            }]
        })

    def set_organization_group_properties(self, reg_entity_name, reg_entity_id, reg_entity_type, properties):
        log.info("Organization Group Properties defined for resource {0}".format(reg_entity_name))
        self.comms.send(json.dumps(
            self._properties(self.next_id(), reg_entity_id, reg_entity_type,
                             getUTCmillis(), properties)))

    def set_properties(self, reg_entity_obj, properties):
        # RegisteredMetric get parent's resid; RegisteredEntity gets own resid
        reg_entity_id = reg_entity_obj.reg_entity_id

        if isinstance(reg_entity_obj, RegisteredMetric):
            entity = reg_entity_obj.parent.ref_entity
        else:
            entity = reg_entity_obj.ref_entity

        log.info("Properties defined for resource {0}".format(entity.name))
        self.comms.send(json.dumps(
            self._properties(self.next_id(), reg_entity_id, entity.entity_type,
                             getUTCmillis(), properties)))
        if entity.entity_type == "HelixGateway":
            with self.file_ops_lock:
                self.store_reg_entity_attributes("EdgeSystem", entity.name,
                                                 reg_entity_obj.reg_entity_id, None, properties)
        else:
            # get dev_type, and prop_dict if possible
            with self.file_ops_lock:
                self.store_reg_entity_attributes("Devices", entity.name, reg_entity_obj.reg_entity_id,
                                                 entity.entity_type, properties)

    def publish_unit(self, reg_entity_obj, metric_name, unit):
        str_prefix, str_unit_name = parse_unit(unit)
        if not isinstance(str_prefix, basestring):
            str_prefix = ""
        if not isinstance(str_unit_name, basestring):
            str_unit_name = ""
        properties_added = {
            metric_name + "_unit": str_unit_name,
            metric_name + "_prefix": str_prefix
        }
        self.set_properties(reg_entity_obj, properties_added)
        log.info("Published metric unit with prefix to IoTCC")

    def prettify(self, elem):
        """Return a pretty-printed XML string for the Element.
        """
        rough_string = ET.tostring(elem)
        reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent="    ")

    def _create_iotcc_json(self):
        msg = {
            "iotcc": {
                "EdgeSystem": {"SystemName": "", "EntityType": "", "uuid": ""},
                "OGProperties": {"OrganizationGroup": ""},
                "Devices": []
            }
        }

        iotcc_path = read_liota_config('IOTCC_PATH', 'iotcc_path')
        path = os.path.dirname(iotcc_path)
        mkdir(path)
        try:
            with open(iotcc_path, 'w') as f:
                json.dump(msg, f, sort_keys=True, indent=4, ensure_ascii=False)
                log.debug('Initialized ' + iotcc_path)
            f.close()
        except IOError, err:
            log.error('Could not open {0} file '.format(iotcc_path) + err)
        return iotcc_path

    def store_reg_entity_details(self, entity_type, entity_name, reg_entity_id):
        msg = ''
        if self._iotcc_json == '':
            log.warn('iotcc.json file missing')
            return
        try:
            with open(self._iotcc_json, 'r') as f:
                msg = json.load(f)
            f.close()
        except IOError, err:
            log.error('Could not open {0} file '.format(self._iotcc_json) + str(err))
        log.debug('{0}:{1}'.format(entity_name, reg_entity_id))
        if entity_type == "HelixGateway":
            msg["iotcc"]["EdgeSystem"]["SystemName"] = entity_name
            msg["iotcc"]["EdgeSystem"]["uuid"] = reg_entity_id
            msg["iotcc"]["EdgeSystem"]["EntityType"] = entity_type
        else:
            entity_exist = False
            for device in msg["iotcc"]["Devices"]:
                if device["uuid"] == reg_entity_id and device["EntityType"] == entity_type and device[
                    "uuid"] == reg_entity_id:
                    entity_exist = True
                    break
            if not entity_exist:
                msg["iotcc"]["Devices"].append(
                    {"DeviceName": entity_name, "uuid": reg_entity_id, "EntityType": entity_type})
        if msg != '':
            with open(self._iotcc_json, 'w') as f:
                json.dump(msg, f, sort_keys=True, indent=4, ensure_ascii=False)
            f.close()

    def remove_reg_entity_details(self, entity_name, reg_entity_id):
        if self._iotcc_json == '':
            log.warn('iotcc.json file missing')
            return
        try:
            with open(self._iotcc_json, 'r') as f:
                msg = json.load(f)
            f.close()
        except IOError, err:
            log.error('Could not open {0} file '.format(self._iotcc_json) + str(err))
        log.debug('Remove {0}:{1} from iotcc.json'.format(entity_name, reg_entity_id))
        if msg["iotcc"]["EdgeSystem"]["SystemName"]==entity_name and msg["iotcc"]["EdgeSystem"]["uuid"]==reg_entity_id:
            del msg["iotcc"]["EdgeSystem"]
            log.info("Removed {0} edge-system from iotcc.json".format(entity_name))
        else:
            entity_exist = False
            for device in msg["iotcc"]["Devices"]:
                if device["uuid"] == reg_entity_id and device["uuid"] == reg_entity_id :
                    entity_exist = True
                    try:
                        msg["iotcc"]["Devices"].remove(device)
                        log.info("Device {0} removed from iotcc.json".format(entity_name))
                        break
                    except ValueError:
                        pass
            if not entity_exist:
                log.error("No such device {0} exists".format(entity_name))
        with open(self._iotcc_json, 'w') as f:
            json.dump(msg, f, sort_keys=True, indent=4, ensure_ascii=False)
        f.close()

    def store_edge_system_info(self, uuid, name, prop_dict):
        """
        create (can overwrite) edge system info file of UUID.xml, with format of
        <attributes>
        <attribute name=attribute name value=attribute value/>
        …
        </attributes>
        except the first attribute is edge system name, all other attributes may vary
        """

        log.debug("store_edge_system_info")
        log.debug('{0}:{1}, prop_list: {2}'.format(uuid, name, prop_dict))
        root = ET.Element("attributes")
        # add edge system name as an attribute
        ET.SubElement(root, "attribute", name="edge system name", value=name)
        # add edge system properties as attributes
        if prop_dict is not None:
            for key in prop_dict.iterkeys():
                value = prop_dict[key]
                if key == 'entity type' or key == 'name' or key == 'device type':
                    continue
                ET.SubElement(root, "attribute", name=key, value=value)
        # add time stamp
        ET.SubElement(root, "attribute", name="LastSeenTimestamp",
                      value=strftime("%Y-%m-%dT%H:%M:%S", gmtime()))

        log.debug("store_edge_system_info dev_file_path:{0}".format(self.dev_file_path))
        file_path = self.dev_file_path + '/' + uuid + '.xml'
        with open(file_path, "w") as fp:
            fp.write(self.prettify(root))
        return

    def store_device_info(self, uuid, name, dev_type, prop_dict, remove_device):
        """
        create (can overwrite) device info file of device_UUID.json, with format of
        {
            "discovery":  {
                "remove" : false,
                "attributes": [
                    {"IoTDeviceType" : "LM35"},
                    {"IoTDeviceName" : "LM35-12345"},
                    {"model": "LM35-A2"},
                    {"function": "thermistor"},
                    {"port" : "GPIO-3"},
                    {"manufacturer" : "Texas Instrument"},
                    {"LastSeenTimestamp" : "04 NOV 2016"}
                ]
            }
        }
        except IoTDeviceType and IoTDeviceName, all other attributes may vary
        """
        log.debug("store_device_info")
        log.debug('prop_dict: {0}'.format(prop_dict))
        attribute_list = [{"IoTDeviceType": dev_type},
                          {"IoTDeviceName": name}]
        # attribute_list.append(prop_dict)
        if prop_dict is not None:
            for key in prop_dict.iterkeys():
                value = prop_dict[key]
                if key == 'entity type' or key == 'name' or key == 'device type':
                    continue
                attribute_list.append({key: value})
        attribute_list.append({"LastSeenTimestamp": strftime("%Y-%m-%dT%H:%M:%S", gmtime())})
        log.debug('attribute_list: {0}'.format(attribute_list))
        msg = {
            "discovery": {
                "remove": remove_device,
                "attributes": attribute_list
            }
        }
        log.debug('msg: {0}'.format(msg))
        log.debug("store_device_info dev_file_path:{0}".format(self.dev_file_path))
        file_path = self.dev_file_path + '/' + uuid + '.json'
        try:
            with open(file_path, 'w') as f:
                json.dump(msg, f, sort_keys=True, indent=4, ensure_ascii=False)
                log.debug('Initialized ' + file_path)
            f.close()
        except IOError, err:
            log.error('Could not open {0} file '.format(file_path) + err)

    def write_entity_file(self, prop_dict, res_uuid):
        file_path = self.entity_file_path + '/' + res_uuid + '.json'
        try:
            with open(file_path, "w") as json_file:
                if (prop_dict is not None):
                    json_string = json.dumps(prop_dict)
                    json_file.write(json_string)
        except:
            log.error('Write file error')

    def read_entity_file(self, res_uuid):
        file_path = self.entity_file_path + '/' + res_uuid + '.json'
        prop_dict = None
        try:
            with open(file_path, "r") as json_file:
                prop_dict = json.loads(json_file.read())
        except:
            log.error('Read file error')
        return prop_dict

    def store_reg_entity_attributes(self, entity_type, entity_name, reg_entity_id,
                                    dev_type, prop_dict):
        log.debug('store_reg_entity_attributes\n {0}:{1}:{2}:{3}'.format(entity_type,
                                                                         entity_name, reg_entity_id, prop_dict))

        ### Update IOTCC local entity file first
        # look for uuid.json file first, if not, first time to write
        # name + type (edge system or device) + if device, device type, + prop_dict
        # if file exists, check name, type and device type match or not
        # if match, merge prop_dict to that file (with existing properties)
        # if not match, replace with new above info + only prop_dict
        # (old property will not be used since outdated already)
        file_path = self.entity_file_path + '/' + reg_entity_id + '.json'
        if not os.path.exists(file_path):
            tmp_dict = {'entity type': str(entity_type), 'name': str(entity_name)}
            if (dev_type is not None):
                tmp_dict.update({"device type": str(dev_type)})
            else:
                tmp_dict.update({"device type": ""})
            if (tmp_dict is not None) and (prop_dict is not None):
                new_prop_dict = dict(tmp_dict.items() + prop_dict.items())
            else:
                new_prop_dict = tmp_dict
        else:
            tmp_dict = self.read_entity_file(reg_entity_id)
            if ((tmp_dict["entity type"] == entity_type) and (tmp_dict["name"] == entity_name)
                and (tmp_dict["device type"] == dev_type)):
                # the same entity
                if (tmp_dict is not None) and (prop_dict is not None):
                    new_prop_dict = dict(tmp_dict.items() + prop_dict.items())
                else:
                    new_prop_dict = tmp_dict
            else:
                tmp_dict = {'entity type': str(entity_type), 'name': str(entity_name)}
                if (dev_type is not None):
                    tmp_dict.update({"device type": str(dev_type)})
                else:
                    tmp_dict.update({"device type": ""})
                if (tmp_dict is not None) and (prop_dict is not None):
                    new_prop_dict = dict(tmp_dict.items() + prop_dict.items())
                else:
                    new_prop_dict = tmp_dict

        # write new property dictionary to local entity file
        self.write_entity_file(new_prop_dict, reg_entity_id)
        ### Write IOTCC device file for AW agents
        if entity_type == "EdgeSystem":
            self.store_edge_system_info(reg_entity_id, entity_name, new_prop_dict)
        elif entity_type == "Devices":
            self.store_device_info(reg_entity_id, entity_name, dev_type, new_prop_dict, False)
        else:
            return

    def _get_file_storage_path(self, name):
        log.debug("_get_{0}".format(name))
        config = ConfigParser.RawConfigParser()
        fullPath = LiotaConfigPath().get_liota_fullpath()
        if fullPath != '':
            try:
                if config.read(fullPath) != []:
                    try:
                        # retrieve device info file storage directory
                        file_path = config.get('IOTCC_PATH', name)
                        log.debug("_get_{0} file_path:{1}".format(name, file_path))
                    except ConfigParser.ParsingError as err:
                        log.error('Could not open config file ' + err)
                        return None
                    if not os.path.exists(file_path):
                        try:
                            os.makedirs(file_path)
                        except OSError as exc:  # Python >2.5
                            if exc.errno == errno.EEXIST and os.path.isdir(file_path):
                                pass
                            else:
                                log.error('Could not create file storage directory')
                                return None
                    log.debug("_get_{0} file_path:{1}".format(name, file_path))
                    return file_path
                else:
                    log.error('Could not open config file ' + fullPath)
                    return None
            except IOError, err:
                log.error('Could not open config file')
                return None
        else:
            # missing config file
            log.warn('liota.conf file missing')
            return None

    def next_id(self):
        self.counter = (self.counter + 1) & 0xffffff
        # Enforce even IDs
        return int(self.counter * 2)

    def _unregistration(self, msg_id, uuid):
        return {
            "transactionID": msg_id,
            "type": "remove_resource_request",
            "body": {
                "uuid": uuid
            }
        }



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

""" Utility functions required for initialization of various objects

"""

from datetime import datetime
import hashlib
import logging
import os
import platform
import random
import uuid
import errno
import ConfigParser
import stat
import json
import subprocess


log = logging.getLogger(__name__)


class systemUUID:
    __UUID = ''

    def __init__(self):
        if systemUUID.__UUID == '':
            self._set_system_uuid()

    # Using uuid.getnode() in order to get the mac address. If all attempts to obtain the hardware address fail, this function returns a random 48-bit number
    # with its eighth bit set to 1 as recommended in RFC 4122. A check on the
    # 8th bit is done in order to figure out if random MAC Address generation
    # is required or not.
    def _getMacAddrIfaceHash(self):
        mac = uuid.getnode()
        if (mac >> 40) % 2:
            log.warn(
                'could not find a mac address, an unlikely potential exists for uuid collisions with liota instances on other IoT gateways')
            # generate a 48-bit random integer from this seed
            # always returns the same random integer
            # however in get_uuid below a unique uuid for each resource name will be created
            # this allows us not to have to store any uuid on the persistent storage yet
            # create a unique system uuid
            random.seed(1234567)
            mac = random.randint(0, 281474976710655)
        m = hashlib.md5()
        m.update(str(mac))
        self.macHash = m.hexdigest()
        return self.macHash

    def _set_system_uuid(self):
        # start by creating the liota namespace, this is a globally unique
        # uuid, and exactly the same for any instance of liota
        self.liotaNamespace = uuid.uuid5(
            uuid.NAMESPACE_URL, 'https://github.com/vmware/liota')
        log.debug(str('liota namespace uuid: ' + str(self.liotaNamespace)))
        # we create a system uuid for the physical system on which this instance is running
        # we hash the interface name with the mac address in getMacAddrIfaceHash to avoid collision of
        # mac addresses across potential different physical interfaces in the
        # IoT space
        systemUUID.__UUID = uuid.uuid5(
            self.liotaNamespace, self._getMacAddrIfaceHash())
        log.debug('system UUID: ' + str(systemUUID.__UUID))

    def _get_system_uuid(self):
        if systemUUID.__UUID == '':
            self._set_system_uuid()
        return systemUUID.__UUID

    def get_uuid(self, resource_name):
        # creating a uuid for a particular resource on this system
        # all resources with the same resource name will have the same uuid, on this system
        # resources created on a different system with a name used on this
        # system will have *different* uuids' as it should be
        tmp = str(uuid.uuid5(self._get_system_uuid(), resource_name.encode('utf-8')))
        log.info('resource: ' + resource_name + '  uuid=' + str(tmp))
        return str(tmp)


def get_linux_version():
    return platform.platform()


def get_default_network_interface():
    """
    Works with Linux.
    There are situations where route may not actually return a default route in the
    main routing table, as the default route might be kept in another table.
    Such cases should be handled manually.
    :return: Default Network Interface of the Edge_System
    """
    cmd = "route | grep '^default' | grep -o '[^ ]*$'"
    nw_iface = str(subprocess.check_output(cmd, shell=True)).rstrip()
    log.info("Default network interface is : {0}".format(nw_iface))
    return nw_iface


def get_disk_name():
    """
    Works with Linux.
    If edge_system has multiple disks, only first disk will be returned.
    Such cases should be handled manually.

    :return: Disk type of the Edge_System
    """
    cmd = "lsblk -io KNAME,TYPE | grep 'disk' | sed -n '1p' | grep -o '^\S*'"
    disk_name = str(subprocess.check_output(cmd, shell=True)).rstrip()
    log.info("Disk name is : {0}".format(disk_name))
    return disk_name


def getUTCmillis():
    return long(1000 * ((datetime.utcnow() - datetime(1970, 1, 1)).total_seconds()))

  
def mkdir(path):
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise


def store_edge_system_uuid(entity_name, entity_id, reg_entity_id):
    """
    Utility function to store EdgeSystem's Name, local-uuid and registered-uuid in the
    specified file.
    :param entity_name: EdgeSystem's Name
    :param entity_id: Local uuid of the EdgeSystem
    :param reg_entity_id: Registered uuid of the EdgeSystem
    :return: None
    """
    try:
        uuid_path = read_liota_config('UUID_PATH', 'uuid_path')
        uuid_config = ConfigParser.RawConfigParser()
        uuid_config.optionxform = str
        uuid_config.add_section('GATEWAY')
        uuid_config.set('GATEWAY', 'name', entity_name)
        if entity_id:
            uuid_config.set('GATEWAY', 'local-uuid', entity_id)
        if reg_entity_id:
            uuid_config.set('GATEWAY', 'registered-uuid', reg_entity_id)
        with open(uuid_path, 'w') as configfile:
            uuid_config.write(configfile)
    except ConfigParser.ParsingError, err:
        log.error('Could not open config file ' + str(err))


class LiotaConfigPath:
    path_liota_config = ''
    syswide_path = '/etc/liota/conf/'

    def __init__(self):
        if LiotaConfigPath.path_liota_config == '':
            self._find_path()

    def _find_path(self):
        """ a multi-step search for the configuration file.
        1. Current working directory. ./liota.conf.
        2. User's home directory (~user/liota.conf)
        3. A place named by an environment variable (LIOTA_CONF)
        4. A standard system-wide directory (such as /etc/liota/conf/liota.conf)
        # assert: every install will have a default liota.conf in /etc/liota/conf
        """

        for loc in os.curdir, os.path.expanduser("~"), os.environ.get(
                "LIOTA_CONF"), LiotaConfigPath.syswide_path:
            if loc is None:
                continue
            path = os.path.join(loc, 'liota.conf')
            if os.path.exists(path):
                LiotaConfigPath.path_liota_config = path
                break
            else:
                continue
        if LiotaConfigPath.path_liota_config == '':
            log.error('liota.conf file not found')

    def get_liota_fullpath(self):
        return LiotaConfigPath.path_liota_config

    def setup_logging(self, default_level=logging.WARNING):
        """
        Setup logging configuration
        """
        log = logging.getLogger(__name__)
        config = ConfigParser.RawConfigParser()
        fullPath = self.get_liota_fullpath()
        if fullPath != '':
            try:
                if config.read(fullPath) != []:
                    # now use json file for logging settings
                    try:
                        log_path = config.get('LOG_PATH', 'log_path')
                        log_cfg = config.get('LOG_CFG', 'json_path')
                    except ConfigParser.ParsingError as err:
                        log.error('Could not parse log config file')
                else:
                    raise IOError('Cannot open configuration file ' + fullPath)
            except IOError as err:
                log.error('Could not open log config file')
            mkdir(log_path)
            if os.path.exists(log_cfg):
                with open(log_cfg, 'rt') as f:
                    config = json.load(f)
                logging.config.dictConfig(config)
                log.info('created logger with ' + log_cfg)
            else:
                # missing logging.json file
                logging.basicConfig(level=default_level)
                log.warn(
                    'logging.json file missing,created default logger with level = ' +
                    str(default_level))
        else:
            # missing config file
            log.warn('liota.conf file missing')

            
def read_liota_config(section, name):
    """
    Returns the value of name within the specified section.
    """
    config = ConfigParser.RawConfigParser()
    fullPath = LiotaConfigPath().get_liota_fullpath()
    if fullPath != '':
        try:
            if config.read(fullPath) != []:
                try:
                    value = config.get(section, name)
                except ConfigParser.ParsingError as err:
                    log.error('Could not parse log config file' + str(err))
            else:
                raise IOError('Cannot open configuration file ' + fullPath)
        except IOError as err:
            log.error('Could not open log config file')
    else:
        # missing config file
        log.warn('liota.conf file missing')
    return value


class DiscUtilities:
    """
    DiscUtilities is a wrapper of utility functions
    """

    def __init__(self):
        pass

    def validate_named_pipe(self, pipe_file):
        assert (isinstance(pipe_file, basestring))
        if os.path.exists(pipe_file):
            if stat.S_ISFIFO(os.stat(pipe_file).st_mode):
                pass
            else:
                log.error("Pipe path exists, but it is not a pipe")
                return False
        else:
            pipe_dir = os.path.dirname(pipe_file)
            if not os.path.isdir(pipe_dir):
                try:
                    os.makedirs(pipe_dir)
                    log.info("Created directory: " + pipe_dir)
                except OSError:
                    log.error("Could not create directory for messenger pipe")
                    return False
            try:
                os.mkfifo(pipe_file)
                log.info("Created pipe: " + pipe_file)
            except OSError:
                log.error("Could not create messenger pipe")
                return False
        assert (stat.S_ISFIFO(os.stat(pipe_file).st_mode))
        return True



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from Queue import Queue
import inspect
import logging
from liota.core import metric_handler
from liota.entities.registered_entity import RegisteredEntity
from liota.lib.utilities.utility import getUTCmillis


log = logging.getLogger(__name__)


class RegisteredMetric(RegisteredEntity):

    def __init__(self, ref_metric, ref_dcc, reg_entity_id):
        super(RegisteredMetric, self).__init__(ref_entity=ref_metric,
                                  ref_dcc=ref_dcc,
                                  reg_entity_id=reg_entity_id)
        self.flag_alive = False
        self._next_run_time = None
        self.current_aggregation_size = 0
        # -------------------------------------------------------------------
        # Elements in this queue are (ts, v) pairs.
        #
        self.values = Queue()

    def start_collecting(self):
        self.flag_alive = True
        # TODO: Add a check to ensure that start_collecting for a metric is
        # called only once by the client code
        metric_handler.initialize()
        self._next_run_time = getUTCmillis() + (self.ref_entity.interval * 1000)
        metric_handler.event_ds.put_and_notify(self)

    def stop_collecting(self):
        self.flag_alive = False
        log.debug("Metric %s is marked for deletion" %
                 str(self.ref_entity.name))

    def add_collected_data(self, collected_data):
        if isinstance(collected_data, list):
            for data_sample in collected_data:
                self.values.put(data_sample)
            return len(collected_data)
        elif isinstance(collected_data, tuple):
            self.values.put(collected_data)
            return 1
        else:
            self.values.put((getUTCmillis(), collected_data))
            return 1

    def get_next_run_time(self):
        return self._next_run_time

    def set_next_run_time(self):
        self._next_run_time = self._next_run_time + \
            (self.ref_entity.interval * 1000)
        log.debug("Set next run time to:" + str(self._next_run_time))

    def is_ready_to_send(self):
        log.debug("self.current_aggregation_size:" +
                  str(self.current_aggregation_size))
        log.debug("self.aggregation_size:" +
                  str(self.ref_entity.aggregation_size))
        return self.current_aggregation_size >= self.ref_entity.aggregation_size

    def collect(self):
        log.debug("Collecting values for the resource {0} ".format(
            self.ref_entity.name))
        self.args_required = len(inspect.getargspec(
            self.ref_entity.sampling_function)[0])
        if self.args_required is not 0:
            self.collected_data = self.ref_entity.sampling_function(1)
        else:
            self.collected_data = self.ref_entity.sampling_function()
        log.debug("Size of the queue {0}".format(self.values.qsize()))
        #  Sampling function might return 'None' because of filtering
        if self.collected_data is not None:
            log.info("{0} Sample Value: {1}".format(
                self.ref_entity.name, self.collected_data))
            no_of_values_added = self.add_collected_data(self.collected_data)
            self.current_aggregation_size = self.current_aggregation_size + no_of_values_added

    def reset_aggregation_size(self):
        self.current_aggregation_size = 0

    def send_data(self):
        log.info("Publishing values for the resource {0} ".format(
            self.ref_entity.name))
        if not self.values:
            # No values measured since last report_data
            return True
        self.ref_dcc.publish(self)

    def __str__(self, *args, **kwargs):
        return str(self.ref_entity.name) + ":" + str(self._next_run_time)

    def __cmp__(self, other):
        if other is None:
            return -1
        if not isinstance(other, RegisteredMetric):
            return -1
        return cmp(self._next_run_time, other._next_run_time)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from Queue import Queue, PriorityQueue, Full
import heapq
import logging
from threading import Thread, Condition, Lock
from time import time as _time

from liota.lib.utilities.utility import getUTCmillis
from liota.lib.utilities.utility import read_liota_config

log = logging.getLogger(__name__)

event_ds = None
collect_queue = None
send_queue = None
event_checker_thread = None
send_thread = None
collect_thread_pool = None


class EventsPriorityQueue(PriorityQueue):

    def __init__(self):
        PriorityQueue.__init__(self)
        self.first_element_changed = Condition(self.mutex)

    def put_and_notify(self, item, block=True, timeout=None):
        log.debug("Adding Event:" + str(item))
        self.not_full.acquire()
        try:
            first_element_before_insertion = None
            if self._qsize() > 0:
                first_element_before_insertion = heapq.nsmallest(1, self.queue)[
                    0]

            if self.maxsize > 0:
                if not block:
                    if self._qsize() == self.maxsize:
                        raise Full
                elif timeout is None:
                    while self._qsize() == self.maxsize:
                        self.not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    endtime = _time() + timeout
                    while self._qsize() == self.maxsize:
                        remaining = endtime - _time()
                        if remaining <= 0.0:
                            raise Full
                        self.not_full.wait(remaining)
            self._put(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()

            first_element_after_insertion = heapq.nsmallest(1, self.queue)[0]
            if first_element_before_insertion != first_element_after_insertion:
                self.first_element_changed.notify()
        finally:
            self.not_full.release()

    def get_next_element_when_ready(self):
        self.first_element_changed.acquire()
        try:
            isNotReady = True
            while isNotReady:
                if self._qsize() > 0:
                    first_element = heapq.nsmallest(1, self.queue)[0]
                    if isinstance(first_element, SystemExit):
                        first_element = self._get()
                        break
                    if not first_element.flag_alive:
                        log.debug("Early termination of dead metric")
                        first_element = self._get()
                        break
                    timeout = (
                        first_element.get_next_run_time() - getUTCmillis()
                    ) / 1000.0
                    log.debug("Waiting on acquired first_element_changed LOCK "
                             + "for: %.2f" % timeout)
                    self.first_element_changed.wait(timeout)
                else:
                    self.first_element_changed.wait()
                    first_element = heapq.nsmallest(1, self.queue)[0]
                if isinstance(first_element, SystemExit):
                    first_element = self._get()
                    break
                if (first_element.get_next_run_time() - getUTCmillis()) <= 0 \
                        or not first_element.flag_alive:
                    isNotReady = False
                    first_element = self._get()
            return first_element
        finally:
            self.first_element_changed.release()


class EventCheckerThread(Thread):

    def __init__(self, name=None):
        Thread.__init__(self, name=name)
        self.flag_alive = True
        self.start()

    def run(self):
        log.info("Started EventCheckerThread")
        global event_ds
        global collect_queue
        while self.flag_alive:
            log.debug("Waiting for event...")
            metric = event_ds.get_next_element_when_ready()
            if isinstance(metric, SystemExit):
                log.debug("Got exit signal")
                break
            log.debug("Got event:" + str(metric))
            if not metric.flag_alive:
                log.debug("Discarded dead metric: %s" % str(metric))
                continue
            collect_queue.put(metric)
        log.info("Thread exits: %s" % str(self.name))


class SendThread(Thread):

    def __init__(self, name=None):
        Thread.__init__(self, name=name)
        self.flag_alive = True
        self.start()

    def run(self):
        log.info("Started SendThread")
        global send_queue
        while self.flag_alive:
            log.debug("Waiting to send...")
            metric = send_queue.get()
            if isinstance(metric, SystemExit):
                log.debug("Got exit signal")
                break
            log.debug("Got item in send_queue: " + str(metric))
            if not metric.flag_alive:
                log.debug("Discarded dead metric: %s" % str(metric))
                continue
            metric.send_data()
        log.info("Thread exits: %s" % str(self.name))


class CollectionThread(Thread):

    def __init__(self, worker_stat_lock, name=None):
        Thread.__init__(self, name=name)
        self.daemon = True
        self.working_obj = None
        self._worker_stat_lock = worker_stat_lock
        self.start()

    def run(self):
        global event_ds
        global collect_queue
        global send_queue
        while True:
            metric = collect_queue.get()
            log.debug("Collecting stats for metric: " + str(metric))
            try:
                if not metric.flag_alive:
                    log.debug("Discarded dead metric: %s" % str(metric))
                    continue
                with self._worker_stat_lock:
                    self.working_obj = metric
                metric.collect()
                with self._worker_stat_lock:
                    self.working_obj = None
                if not metric.flag_alive:
                    log.debug("Discarded dead metric: %s" % str(metric))
                    continue
                metric.set_next_run_time()
                event_ds.put_and_notify(metric)
                if metric.is_ready_to_send():
                    send_queue.put(metric)
                    metric.reset_aggregation_size()
            except Exception as e:
                log.error("Error collecting data for metric" + str(metric))
                raise e


class CollectionThreadPool:

    def __init__(self, num_threads):
        self._num_threads = num_threads
        self._pool = []
        self._worker_stat_lock = Lock()

        log.info("Starting " + str(num_threads) + " for collection")
        for j in range(num_threads):
            self._pool.append(CollectionThread(
                self._worker_stat_lock,
                name="Collector-%d" % (j + 1)
            ))

    def get_num_threads(self):
        return self._num_threads

    def get_stats_working(self):
        num_working = 0
        num_alive = 0
        num_all = 0
        with self._worker_stat_lock:
            for tref in self._pool:
                if not isinstance(tref, Thread):
                    continue
                num_all += 1
                if tref.isAlive():
                    num_alive += 1
                if not tref.working_obj is None:
                    num_working += 1
        return [num_working,
                num_alive,
                num_all,
                self._num_threads]

is_initialization_done = False


def initialize():
    global is_initialization_done
    if is_initialization_done:
        log.debug("Initialization already done")
        pass
    else:
        log.debug("Initializing.............")
        global event_ds
        if event_ds is None:
            event_ds = EventsPriorityQueue()
        global event_checker_thread
        if event_checker_thread is None:
            event_checker_thread = EventCheckerThread(
                name="EventCheckerThread")
        global collect_queue
        if collect_queue is None:
            collect_queue = Queue()
        global send_queue
        if send_queue is None:
            send_queue = Queue()
        global send_thread
        if send_thread is None:
            send_thread = SendThread(name="SendThread")
        global collect_thread_pool
        collect_thread_pool_size = int(read_liota_config('CORE_CFG','collect_thread_pool_size')) 
        collect_thread_pool = CollectionThreadPool(collect_thread_pool_size)
        is_initialization_done = True


def terminate():
    global event_checker_thread
    if event_checker_thread:
        event_checker_thread.flag_alive = False
    global send_thread
    if send_thread:
        send_thread.flag_alive = False
    global event_ds
    if event_ds:
        event_ds.put_and_notify(SystemExit(), timeout=0)
    global send_queue
    if send_queue:
        send_queue.put(SystemExit())



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from liota.core.package_manager import LiotaPackage

dependencies = ["iotcc", "examples/thermistor_simulator"]


def static_vars(**kwargs):
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate


class PackageClass(LiotaPackage):

    def create_udm(self, thermistor_model):
        ureg = thermistor_model.ureg

        import math

        #-------------------------------------------------------------------
        # The following functions operate on physical variables represented
        # in pint objects, and returns a pint object, too.
        # Decorators provided by the pint library are used to check the
        # dimensions of arguments passed to the functions.

        @ureg.check(ureg.volt, ureg.ohm, ureg.volt)
        def get_rx(u, r0, ux):
            rx = r0 * ux / (u - ux)
            return rx

        @ureg.check(
            1 / ureg.kelvin,
            1 / ureg.kelvin,
            1 / ureg.kelvin,
            ureg.ohm)
        def get_temperature(c1, c2, c3, rx):
            temper = 1 / (
                c1 +
                c2 * math.log(rx / ureg.ohm) +
                c3 * math.log(rx / ureg.ohm) ** 3
            )

            #---------------------------------------------------------------
            # Here commented is a counter example, showing how a dimension
            # mismatch can be prevented using pint.
            # Since in the correct one above, the unit of temper is
            # already Kelvin, if we multiply it by ureg.kelvin, the unit of
            # the returned values will become ureg.kelvin ** 2, which will
            # consequently throw an exception in succeeding method calls.

            # temper = 1 / ( \
            #         c1 + \
            #         c2 * math.log(rx / ureg.ohm) + \
            #         c3 * math.log(rx / ureg.ohm) ** 3
            #     ) * ureg.kelvin

            return temper

        #-------------------------------------------------------------------
        # This is a sampling method, which queries the physical model, and
        # calls the physical functions to calculate a desired variable.
        # In this specific case, it gets the coefficients, voltages, and
        # reference resistance from the thermistor simulator, and calls the
        # methods defined above to get the temperature.

        def get_thermistor_temperature():
            temper = get_temperature(
                thermistor_model.get_c1(),
                thermistor_model.get_c2(),
                thermistor_model.get_c3(),
                get_rx(
                    thermistor_model.get_u(),
                    thermistor_model.get_r0(),
                    thermistor_model.get_ux()
                )
            ).to(ureg.degC)
            return temper.magnitude

        self.get_thermistor_temperature = get_thermistor_temperature

    def run(self, registry):
        from liota.entities.metrics.metric import Metric
        import copy

        # Acquire resources from registry
        self.config_path = registry.get("package_conf")
        self.iotcc = registry.get("iotcc")
        self.iotcc_edge_system = copy.copy(registry.get("iotcc_edge_system"))
        thermistor_simulator = registry.get("thermistor_simulator")
        
        self.iotcc_thermistor = self.iotcc.register(thermistor_simulator)
        self.iotcc.create_relationship(self.iotcc_edge_system, self.iotcc_thermistor)

        ureg = thermistor_simulator.ureg
        self.create_udm(thermistor_model=thermistor_simulator)

        # Create metrics
        self.metrics = []
        metric_name = "model.thermistor.temperature"
        thermistor_temper = Metric(
            name=metric_name,
            unit=ureg.degC,
            interval=5,
            sampling_function=self.get_thermistor_temperature
        )
        reg_thermistor_temper = self.iotcc.register(thermistor_temper)
        self.iotcc.create_relationship(self.iotcc_thermistor, reg_thermistor_temper)
        reg_thermistor_temper.start_collecting()
        self.metrics.append(reg_thermistor_temper)

        # Use the iotcc_device_name as identifier in the registry to easily refer the registered device in other packages
        registry.register("iotcc_thermistor_simulated", self.iotcc_thermistor)

    def clean_up(self):

        # Get values from configuration file
        config = {}
        execfile(self.config_path + '/sampleProp.conf', config)

        for metric in self.metrics:
            metric.stop_collecting()

        #Unregister iotcc device
        if config['ShouldUnregisterOnUnload'] == "True":
            self.iotcc.unregister(self.iotcc_thermistor)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#
import logging
from liota.dccs.dcc import DataCenterComponent
from liota.entities.metrics.registered_metric import RegisteredMetric
from liota.entities.metrics.metric import Metric
from liota.entities.registered_entity import RegisteredEntity


log = logging.getLogger(__name__)


class Graphite(DataCenterComponent):
    def __init__(self, comms):
        super(Graphite, self).__init__(
            comms=comms
        )

    def register(self, entity_obj):
        log.info("Registering resource with Graphite DCC {0}".format(entity_obj.name))
        if isinstance(entity_obj, Metric):
            return RegisteredMetric(entity_obj, self, None)
        else:
            return RegisteredEntity(entity_obj, self, None)

    def create_relationship(self, reg_entity_parent, reg_entity_child):
        reg_entity_child.parent = reg_entity_parent

    def _format_data(self, reg_metric):
        met_cnt = reg_metric.values.qsize()
        message = ''
        if met_cnt == 0:
            return
        for _ in range(met_cnt):
            v = reg_metric.values.get(block=True)
            if v is not None:
                # Graphite expects time in seconds, not milliseconds. Hence,
                # dividing by 1000
                message += '%s %s %d\n' % (reg_metric.ref_entity.name,
                                           v[1], v[0] / 1000)
        if message == '':
            return
        log.info ("Publishing values to Graphite DCC")
        log.debug("Formatted message: {0}".format(message))
        return message

    def set_properties(self, reg_entity, properties):
        raise NotImplementedError

    def unregister(self, entity_obj):
        raise NotImplementedError



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import random

from linux_metrics import cpu_stat

from liota.core.package_manager import LiotaPackage

dependencies = ["aws_iot"]


#  Reading CPU Utilization.
def read_cpu_utilization(sample_duration_sec=1):
    cpu_pcts = cpu_stat.cpu_percents(sample_duration_sec)
    return round((100 - cpu_pcts['idle']), 2)


#  Random number generator, simulating living room temperature readings.
def get_living_room_temperature():
    return random.randint(10, 30)


#  Random number generator, simulating living room humidity readings.
def get_living_room_humidity():
    return random.randint(70, 90)


#  Random number generator, simulating living room luminous readings.
def get_living_room_luminance():
    # 0 - Lights Off, 1 - Lights On
    return random.randint(0, 1)

# ------------------------------------------------------------------------------------------------------------------
# In this example, we demonstrate how data for a simulated metric generating
# random numbers can be directed to AWSIoT data center component using Liota.
#
# A Simulated DHT sensor with Temperature and Humidity Metrics and a Simulated
# Digital Light sensor with binary luminance Metrics are used.
#
#
#                                            Dell5kEdgeSystem
#                                                   |
#                                                   |
#                                                   |
#                     -----------------------------------------------------
#                    |                              |                      |
#                    |                              |                      |
#                    |                              |                      |
#                 DHT Sensor                 Digital Light Sensor    CPU Utilization
#                    |                              |                   Metric
#                    |                              |
#            ----------------                  Light Metric
#           |                |
#           |                |
#      Temperature        Humidity
#         Metric           Metric
#
# Developers can leverage the following options:
# ---------------------------------------------
#  a) Use single publish and subscribe topic generated by LIOTA for an EdgeSystem, its Devices and its Metrics.
#  b) Use custom single publish and subscribe topic for an EdgeSystem, its Devices and Metrics.
#
#  - In the above two cases, MQTT message's payload MUST be self-descriptive so that subscriber can subscribe
#    process accordingly to a single topic by parsing payload.
#
#  c) Use custom publish and subscribe topics for Metrics.
#  - In this case, MQTT message's payload need not be self-descriptive.  Subscribers can subscribe to
#    appropriate topics and process accordingly.
#
#  d) Use combination of (a) and (c) or (b) and (c).
#
#
# AWSIoT DCC has enclose_metadata option.  It can be used to enclose EdgeSystem, Device and Metric names
# along with the sensor data payload of a Metric.
#
# This example showcases publishing Metrics using (a) and enclose_metadata
# ----------------------------------------------------------------------------------------------------------------------

class PackageClass(LiotaPackage):

    def run(self, registry):
        import copy
        import pint

        from liota.entities.metrics.metric import Metric
        from liota.entities.devices.simulated_device import SimulatedDevice

        # create a pint unit registry
        ureg = pint.UnitRegistry()

        # Acquire resources from registry
        aws_iot = registry.get("aws_iot")
        aws_iot_edge_system = copy.copy(registry.get("aws_iot_edge_system"))

        # Create metrics
        self.metrics = []

        #  Creating CPU Metric
        cpu_utilization = Metric(
            name="CPUUtilization",
            unit=None,
            interval=10,
            aggregation_size=2,
            sampling_function=read_cpu_utilization
        )
        #  Registering Metric and creating Parent-Child relationship
        reg_cpu_utilization = aws_iot.register(cpu_utilization)
        aws_iot.create_relationship(aws_iot_edge_system, reg_cpu_utilization)
        #  Publishing Registered CPU Utilization Metric to AWSIoT Dcc
        reg_cpu_utilization.start_collecting()
        self.metrics.append(reg_cpu_utilization)

        #  Creating Simulated Device
        dht_sensor = SimulatedDevice("SimulatedDHTSensor")
        #  Registering Device and creating Parent-Child relationship
        reg_dht_sensor = aws_iot.register(dht_sensor)
        aws_iot.create_relationship(aws_iot_edge_system, reg_dht_sensor)
        #  Creating Temperature Metric
        temp_metric = Metric(
            name="LivingRoomTemperature",
            entity_type="Metric",
            unit=ureg.degC,
            interval=1,
            aggregation_size=5,
            sampling_function=get_living_room_temperature
        )
        #  Registering Metric and creating Parent-Child relationship
        reg_temp_metric = aws_iot.register(temp_metric)
        aws_iot.create_relationship(reg_dht_sensor, reg_temp_metric)
        #  Publishing Registered Temperature Metric to AWSIoT Dcc
        reg_temp_metric.start_collecting()
        self.metrics.append(reg_temp_metric)

        #  Creating Humidity Metric
        hum_metric = Metric(
            name="LivingRoomHumidity",
            entity_type="Metric",
            unit=None,
            interval=1,
            aggregation_size=5,
            sampling_function=get_living_room_humidity
        )
        #  Registering Metric and creating Parent-Child relationship
        reg_hum_metric = aws_iot.register(hum_metric)
        aws_iot.create_relationship(reg_dht_sensor, reg_hum_metric)
        #  Publishing Registered Humidity Metric to AWSIoT Dcc
        reg_hum_metric.start_collecting()
        self.metrics.append(reg_hum_metric)

        #  Creating Simulated Device
        light_sensor = SimulatedDevice("SimDigLightSensor")
        #  Registering Device and creating Parent-Child relationship
        reg_light_sensor = aws_iot.register(light_sensor)
        aws_iot.create_relationship(aws_iot_edge_system, reg_light_sensor)

        #  Creating Light Metric
        light_metric = Metric(
            name="LivingRoomLight",
            entity_type="Metric",
            unit=None,
            interval=10,
            aggregation_size=1,
            sampling_function=get_living_room_luminance
        )
        #  Registering Metric and creating Parent-Child relationship
        reg_light_metric = aws_iot.register(light_metric)
        aws_iot.create_relationship(reg_light_sensor, reg_light_metric)
        #  Publishing Registered Light Metric to AWSIoT Dcc
        reg_light_metric.start_collecting()
        self.metrics.append(reg_light_metric)

    def clean_up(self):
        for metric in self.metrics:
            metric.stop_collecting()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#
import logging
import socket

from liota.dcc_comms.dcc_comms import DCCComms


log = logging.getLogger(__name__)


class SocketDccComms(DCCComms):

    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self._connect()

    def _connect(self):
        self.client = socket.socket()
        log.info("Establishing Socket Connection")
        try:
            self.client.connect((self.ip, self.port))
            log.info("Socket Created")
        except Exception as ex:
            log.exception(
                "Unable to establish socket connection. Please check the firewall rules and try again.")
            self.client.close()
            self.client = None
            raise ex

    def _disconnect(self):
        raise NotImplementedError

    def send(self, message, msg_attr=None):
        log.debug("Publishing message:" + str(message))
        if self.client is not None:
            self.client.sendall(message)

    def receive(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import os
import pip
import sys
from pip.req import parse_requirements
from setuptools import setup, find_packages


requirements = [str(requirement.req) for requirement in parse_requirements(
    'requirements.txt', session=pip.download.PipSession())]

# Python Version check
if not sys.version_info[0] == 2:
    sys.exit('Python 3 is not supported')

# Python 2.7.9 sub-version check
if sys.version_info[1] <= 7 and sys.version_info[2] < 9:
    sys.exit('Python 2.7.9+ versions are only supported')

# Get the long description from the README file
with open('README.md') as f:
    long_description = f.read()

setup(
    name='liota',
    version='0.2.6',
    packages=find_packages(exclude=["*.json", "*.txt"]),
    description='IoT Agent',
    long_description=long_description,
    # include_package_data=True

    # The project's main homepage.
    url='https://github.com/vmware/liota',
    author='The Python Packaging Authority',
    author_email='vkohli@vmware.com',

    # License
    license='BSD',
    platforms=['Linux'],

    # Classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        # TO DO: Check for other python versions
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
    ],

    keywords='iot liota agent',

    # Installation requirement
    install_requires=requirements,

    # 'data_file'(conf_files) at custom location
    data_files=[(os.path.abspath(os.sep) + '/../etc/liota/examples',
                 ['examples/simulated_edge_system_graphite.py',
                  'examples/simulated_graphite_event_based.py',
                  'examples/simulated_graphite_temp.py',
                  'examples/dell5k_edge_system_graphite.py',
                  'examples/dell5k_edge_system_iotcc_websocket.py',
                  'examples/dell5k_edge_system_with_filter_iotcc.py',
                  'examples/multi_metric_simulated_edge_system_graphite.py',
                  'examples/dk300_edge_system_iotcc_graphite.py',
                  'examples/sampleProp.conf']),
                (os.path.abspath(os.sep) + '/../etc/liota/examples/model_simulated',
                 ['examples/model_simulated/graphite_bike_simulated.py',
                  'examples/model_simulated/graphite_thermistor_simulated.py',
                  'examples/model_simulated/iotcc_bike_simulated.py',
                  'examples/model_simulated/iotcc_thermistor_simulated.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/examples/mqtt/device_comms/iotcc',
                 ['examples/mqtt/device_comms/iotcc/iotcc_simulated_mqtt.py',
                  'examples/mqtt/device_comms/iotcc/samplePropMqtt.conf']),
                (os.path.abspath(os.sep) + '/../etc/liota/examples/mqtt/dcc_comms/aws_iot',
                 ['examples/mqtt/dcc_comms/aws_iot/sampleProp.conf',
                  'examples/mqtt/dcc_comms/aws_iot/simulated_home_auto_gen_topic.py',
                  'examples/mqtt/dcc_comms/aws_iot/simulated_home_topic_per_metric.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/examples/mqtt/dcc_comms/iotcc',
                 ['examples/mqtt/dcc_comms/iotcc/sampleProp.conf',
                  'examples/mqtt/dcc_comms/iotcc/dell5k_edge_system_iotcc_mqtt.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/packages',
                 ['packages/aws_iot.py',
                  'packages/graphite.py',
                  'packages/iotcc.py',
                  'packages/iotcc_mqtt.py',
                  'packages/sampleProp.conf',
                  'packages/liotad.py',
                  'packages/liotapkg.sh',
                  'packages/packages_auto.txt']),
                (os.path.abspath(os.sep) + '/../etc/liota/packages/dev_disc',
                 ['packages/dev_disc/liota_disc_pipe.sh',
                  'packages/dev_disc/dev_disc.py',
                  'packages/dev_disc/liota_devsim_pipe.sh',
                  'packages/dev_disc/liota_devsim_load.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/packages/examples',
                 ['packages/examples/bike_simulator.py',
                  'packages/examples/iotcc_ram.py',
                  'packages/examples/graphite_bike_simulated.py',
                  'packages/examples/iotcc_bike_simulated.py',
                  'packages/examples/graphite_edge_system_stats.py',
                  'packages/examples/graphite_edge_system_stats_with_filter.py',
                  'packages/examples/iotcc_edge_system_stats.py',
                  'packages/examples/iotcc_edge_system_stats_with_filter.py',
                  'packages/examples/iotcc_resource_properties.py',
                  'packages/examples/thermistor_simulator.py',
                  'packages/examples/graphite_thermistor_simulated.py',
                  'packages/examples/iotcc_thermistor_simulated.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/packages/examples/mqtt/aws_iot',
                 ['packages/examples/mqtt/aws_iot/simulated_home_auto_gen_topic.py',
                  'packages/examples/mqtt/aws_iot/simulated_home_topic_per_metric.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/packages/examples/mqtt/iotcc',
                 ['packages/examples/mqtt/iotcc/iotcc_mqtt_edge_system_stats.py',
                  'packages/examples/mqtt/iotcc/iotcc_resource_properties_mqtt.py',
                  'packages/examples/mqtt/iotcc/iotcc_mqtt_ram.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/packages/edge_systems/dell5k',
                 ['packages/edge_systems/dell5k/edge_system.py']),
                (os.path.abspath(os.sep) + '/../etc/liota/scripts',
                 ['scripts/autostartliota']),
                (os.path.abspath(os.sep) + '/../etc/liota/conf',
                 ['config/liota.conf', 'config/logging.json']),
                (os.path.abspath(os.sep) + '/../etc/liota',
                 ['BSD_LICENSE.txt', 'BSD_NOTICE.txt']),
                (os.path.abspath(os.sep) + '/../var/log/liota', [])]
)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import random

import pint
from linux_metrics import cpu_stat

from liota.dccs.aws_iot import AWSIoT
from liota.dcc_comms.mqtt_dcc_comms import MqttDccComms
from liota.entities.metrics.metric import Metric
from liota.entities.devices.simulated_device import SimulatedDevice
from liota.entities.edge_systems.dell5k_edge_system import Dell5KEdgeSystem
from liota.lib.transports.mqtt import QoSDetails
from liota.lib.utilities.identity import Identity
from liota.lib.utilities.tls_conf import TLSConf

# getting aws related values from conf file
config = {}
execfile('sampleProp.conf', config)

# create a pint unit registry
ureg = pint.UnitRegistry()


#  Reading CPU Utilization.
def read_cpu_utilization(sample_duration_sec=1):
    cpu_pcts = cpu_stat.cpu_percents(sample_duration_sec)
    return round((100 - cpu_pcts['idle']), 2)


#  Random number generator, simulating living room temperature readings.
def get_living_room_temperature():
    return random.randint(10, 30)


#  Random number generator, simulating living room humidity readings.
def get_living_room_humidity():
    return random.randint(70, 90)


#  Random number generator, simulating living room luminous readings.
def get_living_room_luminance():
    # 0 - Lights Off, 1 - Lights On
    return random.randint(0, 1)

# ----------------------------------------------------------------------------------------------------------------
# In this example, we demonstrate how data for a simulated metric generating
# random numbers can be directed to AWSIoT data center component using Liota.
#
# A Simulated DHT sensor with Temperature and Humidity Metrics and a Simulated
# Digital Light sensor with binary luminance Metrics are used.
#
#
#                                            Dell5kEdgeSystem
#                                                   |
#                                                   |
#                                                   |
#                     -----------------------------------------------------
#                    |                              |                      |
#                    |                              |                      |
#                    |                              |                      |
#                 DHT Sensor                 Digital Light Sensor    CPU Utilization
#                    |                              |                   Metric
#                    |                              |
#            ----------------                  Light Metric
#           |                |
#           |                |
#      Temperature        Humidity
#         Metric           Metric
#
# Developers can leverage the following options:
# ---------------------------------------------
#  a) Use single publish and subscribe topic generated by LIOTA for an EdgeSystem, its Devices and its Metrics.
#  b) Use custom single publish and subscribe topic for an EdgeSystem, its Devices and Metrics.
#
#  - In the above two cases, MQTT message's payload MUST be self-descriptive so that subscriber can subscribe
#    process accordingly to a single topic by parsing payload.
#
#  c) Use custom publish and subscribe topics for Metrics.
#  - In this case, MQTT message's payload need not be self-descriptive.  Subscribers can subscribe to
#    appropriate topics and process accordingly.
#
#  d) Use combination of (a) and (c) or (b) and (c).
#
#
# AWSIoT DCC has enclose_metadata option.  It can be used to enclose EdgeSystem, Device and Metric names
# along with the sensor data payload of a Metric.
#
# This example showcases publishing Metrics using (a) and enclose_metadata
# -----------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    #  Creating EdgeSystem
    edge_system = Dell5KEdgeSystem(config['EdgeSystemName'])
    #  Encapsulates Identity
    identity = Identity(root_ca_cert=config['broker_root_ca_cert'], username=None, password=None,
                        cert_file=config['edge_system_cert_file'], key_file=config['edge_system_key_file'])
    # Encapsulate TLS parameters
    tls_conf = TLSConf(config['cert_required'], config['tls_version'], config['cipher'])
    # Encapsulate QoS related parameters
    qos_details = QoSDetails(config['in_flight'], config['queue_size'], config['retry'])

    #  Connecting to AWSIoT
    #  AWSIoT broker doesn't support session persistence.  So, always use "clean_session=True"
    #  Publish topic for all Metrics will be 'liota/generated_local_uuid_of_edge_system/request'
    aws = AWSIoT(MqttDccComms(edge_system_name=edge_system.name,
                              url=config['BrokerIP'], port=config['BrokerPort'], identity=identity,
                              tls_conf=tls_conf,
                              qos_details=qos_details,
                              clean_session=True,
                              protocol=config['protocol'], transport=['transport'],
                              conn_disconn_timeout=config['ConnectDisconnectTimeout']),
                 enclose_metadata=True)
    #  Registering EdgeSystem
    reg_edge_system = aws.register(edge_system)

    #  Creating CPU Metric
    cpu_utilization = Metric(
        name="CPUUtilization",
        unit=None,
        interval=10,
        aggregation_size=2,
        sampling_function=read_cpu_utilization
    )
    #  Registering Metric and creating Parent-Child relationship
    reg_cpu_utilization = aws.register(cpu_utilization)
    aws.create_relationship(reg_edge_system, reg_cpu_utilization)
    #  Publishing Registered CPU Utilization Metric to AWS
    reg_cpu_utilization.start_collecting()

    #  Creating Simulated Device
    dht_sensor = SimulatedDevice("SimulatedDHTSensor")
    #  Registering Device and creating Parent-Child relationship
    reg_dht_sensor = aws.register(dht_sensor)
    aws.create_relationship(reg_edge_system, reg_dht_sensor)
    #  Creating Temperature Metric
    temp_metric = Metric(
        name="LivingRoomTemperature",
        entity_type="Metric",
        unit=ureg.degC,
        interval=1,
        aggregation_size=5,
        sampling_function=get_living_room_temperature
    )
    #  Registering Metric and creating Parent-Child relationship
    reg_temp_metric = aws.register(temp_metric)
    aws.create_relationship(reg_dht_sensor, reg_temp_metric)
    #  Publishing Registered Temperature Metric to AWS
    reg_temp_metric.start_collecting()

    #  Creating Humidity Metric
    hum_metric = Metric(
        name="LivingRoomHumidity",
        entity_type="Metric",
        unit=None,
        interval=1,
        aggregation_size=5,
        sampling_function=get_living_room_humidity
    )
    #  Registering Metric and creating Parent-Child relationship
    reg_hum_metric = aws.register(hum_metric)
    aws.create_relationship(reg_dht_sensor, reg_hum_metric)
    #  Publishing Registered Humidity Metric to AWS
    reg_hum_metric.start_collecting()

    #  Creating Simulated Device
    light_sensor = SimulatedDevice("SimDigLightSensor")
    #  Registering Device and creating Parent-Child relationship
    reg_light_sensor = aws.register(light_sensor)
    aws.create_relationship(reg_edge_system, reg_light_sensor)

    #  Creating Light Metric
    light_metric = Metric(
        name="LivingRoomLight",
        entity_type="Metric",
        unit=None,
        interval=10,
        aggregation_size=1,
        sampling_function=get_living_room_luminance
    )
    #  Registering Metric and creating Parent-Child relationship
    reg_light_metric = aws.register(light_metric)
    aws.create_relationship(reg_light_sensor, reg_light_metric)
    #  Publishing Registered Light Metric to AWS
    reg_light_metric.start_collecting()




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import json
import time
import logging

from liota.dev_sims.device_simulator import DeviceSimulator
from liota.device_comms.mqtt_device_comms import MqttDeviceComms
from liota.lib.utilities.identity import Identity
from liota.lib.utilities.tls_conf import TLSConf
from liota.lib.transports.mqtt import QoSDetails


log = logging.getLogger(__name__)


class MqttSimulator(DeviceSimulator):
    """
    MqttSimulator does inter-process communication (IPC), and
    publishes on topics to send simulated device beacon messages.
    """

    def __init__(self, mqtt_cfg, name=None, simulator=None):
        super(MqttSimulator, self).__init__(name=name)
        log.debug("MqttSimulator is initializing")

        broker_ip_port_topic = mqtt_cfg["broker_ip_port_topic"]
        str_list = broker_ip_port_topic.split(':')
        if str_list[0] == "" or str_list[0] == None:
            log.debug("No broker ip is specified!")
            self.broker_ip = "127.0.0.1"
        else:
            self.broker_ip = str(str_list[0])
        if str_list[1] == "" or str_list[1] == None:
            log.debug("No broker port is specified!")
        self.broker_port = int(str_list[1])
        if str_list[2] == "" or str_list[2] == None:
            log.debug("No broker topic is specified!")
        else:
            self.topic = str(str_list[2])
        self.simulator = simulator

        # Encapsulates Identity
        self.identity = Identity(mqtt_cfg['broker_root_ca_cert'], mqtt_cfg['broker_username'],
                                 mqtt_cfg['broker_password'], mqtt_cfg['edge_system_cert_file'],
                                 mqtt_cfg['edge_system_key_file'])
        if ((mqtt_cfg['cert_required'] == "None") or  (mqtt_cfg['cert_required'] == "CERT_NONE")):
            self.tls_conf = None
        else:
            # Encapsulate TLS parameters
            self.tls_conf = TLSConf(mqtt_cfg['cert_required'], mqtt_cfg['tls_version'], mqtt_cfg['cipher'])
        # Encapsulate QoS related parameters
        self.qos_details = QoSDetails(mqtt_cfg['in_flight'], mqtt_cfg['queue_size'], mqtt_cfg['retry'])
        # Create MQTT connection object with required params
        self.mqtt_conn = MqttDeviceComms(url=self.broker_ip, port=int(self.broker_port), identity=self.identity,
                                         tls_conf=self.tls_conf, qos_details=self.qos_details,  clean_session=True,
                                         keep_alive=int(mqtt_cfg['keep_alive']), enable_authentication=False)

        log.debug("MqttSimulator is initialized")
        print "MqttSimulator is initialized"
        self.cfg_sets = mqtt_cfg
        self.cnt = 0
        self.flag_alive = True
        self.start()

    def run(self):
        msg = {
            "Apple56": {
                "k1": "v1",
                "SN": "0",
                "kn": "vn"
            }
        }
        log.info('MqttSimulator is running')
        print "MqttSimulator is running"
        while self.flag_alive:
            if self.cnt >= 5:
                time.sleep(1000);
            else:
                msg["Apple56"]["SN"] = str(self.cnt)
                log.debug("send msg:{0}".format(msg))
                self.mqtt_conn.publish(self.topic, json.dumps(msg), 2, False)
                time.sleep(5)
            self.cnt += 1
            if self.cnt > 20:
                self.flag = False
        log.info("Thread exits: %s" % str(self.name))
        self.mqtt_conn._disconnect()

    def clean_up(self):
        self.flag_alive = False
        self.mqtt_conn._disconnect()